# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import warnings
from test.utils import undeploy_all_version
from time import sleep

import joblib
import merlin
import numpy as np
import pytest
import xgboost as xgb
from merlin.model import ModelType, PyFuncModel
from sklearn import svm
from sklearn.datasets import load_iris

warnings.filterwarnings('ignore')

request_json = {
    "instances": [
        [2.8, 1.0, 6.8, 0.4],
        [3.1, 1.4, 4.5, 1.6]
    ]
}


class EnsembleModel(PyFuncModel):
    def initialize(self, artifacts):
        self._model_1 = xgb.Booster(model_file=artifacts["xgb_model"])
        self._model_2 = joblib.load(artifacts["sklearn_model"])

    def infer(self, model_input):
        inputs = np.array(model_input['instances'])
        dmatrix = xgb.DMatrix(model_input['instances'])
        result_1 = self._model_1.predict(dmatrix)
        result_2 = self._model_2.predict_proba(inputs)
        return {"predictions": ((result_1 + result_2) / 2).tolist()}


class EnvVarModel(PyFuncModel):
    def initialize(self, artifacts):
        self.env_var = {}
        self.env_var["workers"] = os.environ.get("WORKERS")
        self.env_var["env_var_1"] = os.environ.get("ENV_VAR_1")
        self.env_var["env_var_2"] = os.environ.get("ENV_VAR_2")

    def infer(self, model_input):
        return self.env_var

@pytest.mark.pyfunc
@pytest.mark.integration
@pytest.mark.dependency()
def test_pyfunc(integration_test_url, project_name, use_google_oauth, requests):
    merlin.set_url(integration_test_url, use_google_oauth=use_google_oauth)
    merlin.set_project(project_name)
    merlin.set_model("pyfunc-sample", ModelType.PYFUNC)

    undeploy_all_version()
    with merlin.new_model_version() as v:
        iris = load_iris()
        y = iris['target']
        X = iris['data']
        xgb_path = train_xgboost_model(X, y)
        sklearn_path = train_sklearn_model(X, y)

        v.log_pyfunc_model(model_instance=EnsembleModel(),
                           conda_env="test/pyfunc/env.yaml",
                           code_dir=["test"],
                           artifacts={"xgb_model": xgb_path,
                                      "sklearn_model": sklearn_path})

    endpoint = merlin.deploy(v)

    resp = requests.post(f"{endpoint.url}", json=request_json)

    assert resp.status_code == 200
    assert resp.json() is not None
    assert len(resp.json()['predictions']) == len(request_json['instances'])

    merlin.undeploy(v)


@pytest.mark.pyfunc
@pytest.mark.integration
def test_pyfunc_env_vars(integration_test_url, project_name, use_google_oauth, requests):
    merlin.set_url(integration_test_url, use_google_oauth=use_google_oauth)
    merlin.set_project(project_name)
    merlin.set_model("pyfunc-env-vars-sample", ModelType.PYFUNC)

    undeploy_all_version()
    with merlin.new_model_version() as v:
        v.log_pyfunc_model(model_instance=EnvVarModel(),
                           conda_env="test/pyfunc/env.yaml",
                           code_dir=["test"],
                           artifacts={})

    env_vars = {"WORKERS": "8", "ENV_VAR_1": "1", "ENV_VAR_2": "2"}
    endpoint = merlin.deploy(v, env_vars=env_vars)
    resp = requests.post(f"{endpoint.url}", json=request_json)

    assert resp.status_code == 200
    assert resp.json() is not None
    assert resp.json()['workers'] == "8"
    assert resp.json()['env_var_1'] == "1"
    assert resp.json()['env_var_2'] == "2"
    assert env_vars.items() <= endpoint.env_vars.items()

    merlin.undeploy(v)


# This implementation of PyFuncModel uses the old infer method (no keyword arguments).
# The keywords arguments for infer() method introduced in Merlin 0.5.2.
class OldInferModel(PyFuncModel):
    def initialize(self, artifacts):
        pass

    def infer(self, model_input):
        return model_input


@pytest.mark.pyfunc
@pytest.mark.integration
def test_pyfunc_old_infer(integration_test_url, project_name, use_google_oauth, requests):
    merlin.set_url(integration_test_url, use_google_oauth=use_google_oauth)
    merlin.set_project(project_name)
    merlin.set_model("pyfunc-old-infer-sample", ModelType.PYFUNC)

    undeploy_all_version()
    with merlin.new_model_version() as v:
        v.log_pyfunc_model(model_instance=OldInferModel(),
                           conda_env="test/pyfunc/env.yaml",
                           code_dir=["test"],
                           artifacts={})

    endpoint = merlin.deploy(v)
    resp = requests.post(f"{endpoint.url}", json=request_json)

    assert resp.status_code == 200
    assert resp.json()['instances'] == request_json['instances']

    merlin.undeploy(v)


def train_xgboost_model(X, y):
    model_1_dir = "test/pyfunc/"
    BST_FILE = "model_1.bst"
    dtrain = xgb.DMatrix(X, label=y)
    param = {'max_depth': 6,
             'eta': 0.1,
             'silent': 1,
             'nthread': 4,
             'num_class': 3,
             'objective': 'multi:softprob'
             }
    xgb_model = xgb.train(params=param, dtrain=dtrain)
    model_1_path = os.path.join(model_1_dir, BST_FILE)
    xgb_model.save_model(model_1_path)
    return model_1_path


def train_sklearn_model(X, y):
    model_2_dir = "test/pyfunc/"
    MODEL_FILE = "model_2.joblib"
    model_2_path = os.path.join(model_2_dir, MODEL_FILE)

    clf = svm.SVC(gamma='scale', probability=True)
    clf.fit(X, y)
    joblib.dump(clf, model_2_path)
    return model_2_path



# coding: utf-8

"""
    Merlin

    API Guide for accessing Merlin's model management, deployment, and serving functionalities  # noqa: E501

    OpenAPI spec version: 0.14.0
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

import pprint
import re  # noqa: F401

import six

class ResourceRequest(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """
    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'min_replica': 'int',
        'max_replica': 'int',
        'cpu_request': 'str',
        'memory_request': 'str',
        'gpu_name': 'str',
        'gpu_request': 'str'
    }

    attribute_map = {
        'min_replica': 'min_replica',
        'max_replica': 'max_replica',
        'cpu_request': 'cpu_request',
        'memory_request': 'memory_request',
        'gpu_name': 'gpu_name',
        'gpu_request': 'gpu_request'
    }

    def __init__(self, min_replica=None, max_replica=None, cpu_request=None, memory_request=None, gpu_name=None, gpu_request=None):  # noqa: E501
        """ResourceRequest - a model defined in Swagger"""  # noqa: E501
        self._min_replica = None
        self._max_replica = None
        self._cpu_request = None
        self._memory_request = None
        self._gpu_name = None
        self._gpu_request = None
        self.discriminator = None
        if min_replica is not None:
            self.min_replica = min_replica
        if max_replica is not None:
            self.max_replica = max_replica
        if cpu_request is not None:
            self.cpu_request = cpu_request
        if memory_request is not None:
            self.memory_request = memory_request
        if gpu_name is not None:
            self.gpu_name = gpu_name
        if gpu_request is not None:
            self.gpu_request = gpu_request

    @property
    def min_replica(self):
        """Gets the min_replica of this ResourceRequest.  # noqa: E501


        :return: The min_replica of this ResourceRequest.  # noqa: E501
        :rtype: int
        """
        return self._min_replica

    @min_replica.setter
    def min_replica(self, min_replica):
        """Sets the min_replica of this ResourceRequest.


        :param min_replica: The min_replica of this ResourceRequest.  # noqa: E501
        :type: int
        """

        self._min_replica = min_replica

    @property
    def max_replica(self):
        """Gets the max_replica of this ResourceRequest.  # noqa: E501


        :return: The max_replica of this ResourceRequest.  # noqa: E501
        :rtype: int
        """
        return self._max_replica

    @max_replica.setter
    def max_replica(self, max_replica):
        """Sets the max_replica of this ResourceRequest.


        :param max_replica: The max_replica of this ResourceRequest.  # noqa: E501
        :type: int
        """

        self._max_replica = max_replica

    @property
    def cpu_request(self):
        """Gets the cpu_request of this ResourceRequest.  # noqa: E501


        :return: The cpu_request of this ResourceRequest.  # noqa: E501
        :rtype: str
        """
        return self._cpu_request

    @cpu_request.setter
    def cpu_request(self, cpu_request):
        """Sets the cpu_request of this ResourceRequest.


        :param cpu_request: The cpu_request of this ResourceRequest.  # noqa: E501
        :type: str
        """

        self._cpu_request = cpu_request

    @property
    def memory_request(self):
        """Gets the memory_request of this ResourceRequest.  # noqa: E501


        :return: The memory_request of this ResourceRequest.  # noqa: E501
        :rtype: str
        """
        return self._memory_request

    @memory_request.setter
    def memory_request(self, memory_request):
        """Sets the memory_request of this ResourceRequest.


        :param memory_request: The memory_request of this ResourceRequest.  # noqa: E501
        :type: str
        """

        self._memory_request = memory_request

    @property
    def gpu_name(self):
        """Gets the gpu_name of this ResourceRequest.  # noqa: E501


        :return: The gpu_name of this ResourceRequest.  # noqa: E501
        :rtype: str
        """
        return self._gpu_name

    @gpu_name.setter
    def gpu_name(self, gpu_name):
        """Sets the gpu_name of this ResourceRequest.


        :param gpu_name: The gpu_name of this ResourceRequest.  # noqa: E501
        :type: str
        """

        self._gpu_name = gpu_name

    @property
    def gpu_request(self):
        """Gets the gpu_request of this ResourceRequest.  # noqa: E501


        :return: The gpu_request of this ResourceRequest.  # noqa: E501
        :rtype: str
        """
        return self._gpu_request

    @gpu_request.setter
    def gpu_request(self, gpu_request):
        """Sets the gpu_request of this ResourceRequest.


        :param gpu_request: The gpu_request of this ResourceRequest.  # noqa: E501
        :type: str
        """

        self._gpu_request = gpu_request

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(ResourceRequest, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ResourceRequest):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other



# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from merlin.docker.docker import copy_pyfunc_dockerfile, copy_standard_dockerfile


def test_copy_pyfunc_dockerfile():
    path = copy_pyfunc_dockerfile(".")
    assert os.path.isfile(path) == True
    assert os.path.basename(path) == "pyfunc.Dockerfile"


def test_copy_standard_dockerfile():
    path = copy_standard_dockerfile(".")
    assert os.path.isfile(path) == True
    assert os.path.basename(path) == "standard.Dockerfile"



# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click
import warnings
import os
import merlin
from merlin.model import ModelType
from cookiecutter.main import cookiecutter
from merlin.util import valid_name_check

warnings.filterwarnings("ignore")


@click.group()
def cli():
    """
    A simple command line tool.

    The Merlin CLI assumes that you already have a serialized model.\n
    To see the options for each command: merlin COMMAND --help
    """
    pass


@cli.command("deploy", short_help="Deploy the model")
@click.option("--env", "-e", required=True, help="The environment of model deployment")
@click.option("--url", "-u", required=True, help="The endpoint of model deployment")
@click.option(
    "--project", "-p", required=True, help="The project name of model deployment"
)
@click.option(
    "--model-dir", "-m", required=True, help="The directory with model for deployment"
)
@click.option("--model-name", "-n", required=True, help="The model name for deployment")
@click.option(
    "--model-type", "-t", required=True, help="The type of machine learning algorithm"
)
@click.option(
    "--min-replica",
    required=False,
    help="The minimum number of replicas to create for this deployment",
)
@click.option(
    "--max-replica",
    required=False,
    help="The maximum number of replicas to create for this deployment",
)
@click.option(
    "--cpu-request",
    required=False,
    help="The CPU resource requirement requests for this deployment. Example: 100m.",
)
@click.option(
    "--memory-request",
    required=False,
    help="The memory resource requirement requests for this deployment. Example: 256Mi.",
)
def deploy(
    env,
    model_name,
    model_type,
    model_dir,
    project,
    url,
    min_replica,
    max_replica,
    cpu_request,
    memory_request,
):
    merlin.set_url(url)

    target_env = merlin.get_environment(env)

    resource_request = target_env.default_resource_request
    if min_replica is not None:
        resource_request.min_replica = int(min_replica)
    if max_replica is not None:
        resource_request.max_replica = int(max_replica)
    if cpu_request is not None:
        resource_request.cpu_request = cpu_request
    if memory_request is not None:
        resource_request.memory_request = memory_request

    merlin.set_project(project)
    merlin.set_model(model_name, ModelType(model_type))

    with merlin.new_model_version() as v:
        merlin.log_model(model_dir=model_dir)

    try:
        endpoint = merlin.deploy(v, env, resource_request)
        if endpoint:
            print("Model deployed to {}".format(endpoint))
    except Exception as e:
        print(e)


@cli.command("undeploy", short_help="Undeploy the model")
@click.option("--url", "-u", required=True, help="The endpoint of model deployment")
@click.option(
    "--project", "-p", required=True, help="The project name of model deployment"
)
@click.option("--model-name", "-n", required=True, help="The model name for deployment")
@click.option(
    "--model-version", "-v", required=True, help="The model version for deployment"
)
def undeploy(model_name, model_version, project, url):
    merlin.set_url(url)
    merlin.set_project(project)
    merlin.set_model(model_name)

    merlin_active_model = merlin.active_model()
    all_versions = merlin_active_model.list_version()

    try:
        wanted_model_info = [
            model_info
            for model_info in all_versions
            if model_info._id == int(model_version)
        ][0]
    except Exception as e:
        print(e)
        print("Model Version {} is not found.".format(model_version))

    try:
        merlin.undeploy(wanted_model_info)
    except Exception as e:
        print(e)


@cli.command("scaffold", short_help="Generate PyFunc project")
@click.option(
    "--project", "-p", required=True, help="The merlin project name of PyFunc server"
)
@click.option(
    "--model-name",
    "-m",
    required=True,
    help="The model name which will be listed in merlin",
)
@click.option(
    "--env",
    "-e",
    required=True,
    help="The environment which PyFunc server will be deployed, available environment are id and global",
)
def scaffold(project, model_name, env):
    if not valid_name_check(project) or not valid_name_check(model_name):
        print(
            """Your project/model name contains invalid characters.\
                \nUse only the following characters\
                \n- Characters: a-z (Lowercase ONLY)\
                \n- Numbers: 0-9\
                \n- Symbols: -
            """
        )
    else:
        try:
            cookiecutter(
                "git@github.com:caraml-dev/merlin/python/pyfunc-scaffolding",
                checkout="tags/v0.1",
                no_input=True,
                directory="python/pyfunc-scaffolding",
                extra_context={
                    "project_name": project,
                    "model_name": model_name,
                    "environment_name": env,
                },
            )

        except Exception as e:
            print(e)


if __name__ == "__main__":
    cli()



# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from merlin.batch.sink import BigQuerySink, SaveMode


@pytest.mark.parametrize(
    "table,staging_bucket,result_column,save_mode,options,expected_dict,expected_valid", [
        pytest.param(
            'project.dataset.table',
            'bucket',
            'column',
            SaveMode.APPEND,
            {'project': 'project'},
            {
                'table': 'project.dataset.table',
                'staging_bucket': 'bucket',
                'result_column': 'column',
                'save_mode': SaveMode.APPEND.value,
                'options': {'project': 'project'}
            },
            True,
            id="Should return correct dictionary and valid if sink has valid_types and valid property values"
        ),
        pytest.param(
            'project.dataset.table',
            'bucket',
            'column',
            SaveMode.APPEND,
            {},
            {
                'table': 'project.dataset.table',
                'staging_bucket': 'bucket',
                'result_column': 'column',
                'save_mode': SaveMode.APPEND.value,
                'options': {}
            },
            True,
            id="Should return correct dictionary and valid if sink has valid_types and valid property values(2)"
        ),
        pytest.param(
            'project.dataset.table',
            'bucket',
            'column',
            SaveMode.APPEND,
            None,
            {
                'table': 'project.dataset.table',
                'staging_bucket': 'bucket',
                'result_column': 'column',
                'save_mode': SaveMode.APPEND.value,
                'options': {}
            },
            True,
            id="Should return correct dictionary and valid if sink has valid_types and valid property values(3)"
        ),
        pytest.param(
            'project',
            'bucket',
            'column',
            SaveMode.APPEND,
            None,
            {},
            False,
            id="Should return empty dictionary and not valid if table is invalid"
        ),
        pytest.param(
            'project.dataset.table',
            'bucket',
            '_TABLE_abc',
            SaveMode.APPEND,
            None,
            {},
            False,
            id="Should return empty dictionary and not valid if result_column is invalid"
        )
    ]
)
def test_dictionary_generated(table, staging_bucket, result_column, save_mode, options, expected_dict, expected_valid):
    bq_sink = BigQuerySink(table, staging_bucket, result_column, save_mode, options)
    if expected_valid:
        assert bq_sink.to_dict() == expected_dict
    else:
        with pytest.raises(ValueError):
            bq_sink.to_dict()


@pytest.mark.parametrize(
    "table,staging_bucket,result_column,save_mode,options,expected", [
        pytest.param(
            'project.dataset.table',
            'bucket',
            'column',
            SaveMode.APPEND,
            {'project': 'project'},
            True,
            id="Should valid if sink has valid types and valid property values"
        ),
        pytest.param(
            'project.dataset.table',
            1,
            'column',
            SaveMode.APPEND,
            {},
            False,
            id="Should not valid if staging_bucket has wrong data type, it should be string"
        ),
        pytest.param(
            'project.dataset.table',
            'bucket',
            ['column'],
            SaveMode.APPEND,
            {'project': 'project'},
            False,
            id="Should not valid if result_column has wrong data type, it should be string"
        ),
        pytest.param(
            'project.dataset.table',
            'bucket',
            '_PARTITION_result',
            SaveMode.APPEND,
            None,
            False,
            id="Should not valid if result_column is invalid"
        ),
        pytest.param(
            'project.dataset.table',
            'bucket',
            '_TABLE_abc',
            SaveMode.APPEND,
            None,
            False,
            id="Should not valid if result_column is invalid(2)"
        ),
    ]
)
def test_valid(table, staging_bucket, result_column, save_mode, options, expected):
    bq_sink = BigQuerySink(table, staging_bucket, result_column, save_mode, options)
    if expected:
        bq_sink._validate()
    else:
        with pytest.raises(ValueError):
            bq_sink._validate()



# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional


class ResourceRequest:
    """
    The resource requirement and replicas requests for model version endpoint.
    """

    def __init__(self, min_replica: int, max_replica: int, cpu_request: str, memory_request: str, gpu_request: str=None, gpu_name: str=None):
        self._min_replica = min_replica
        self._max_replica = max_replica
        self._cpu_request = cpu_request
        self._memory_request = memory_request
        self._gpu_request = gpu_request
        self._gpu_name = gpu_name
        self.validate()

    @property
    def min_replica(self) -> int:
        return self._min_replica

    @min_replica.setter
    def min_replica(self, min_replica):
        self._min_replica = min_replica

    @property
    def max_replica(self) -> int:
        return self._max_replica

    @max_replica.setter
    def max_replica(self, max_replica):
        self._max_replica = max_replica

    @property
    def cpu_request(self) -> str:
        return self._cpu_request

    @cpu_request.setter
    def cpu_request(self, cpu_request):
        self._cpu_request = cpu_request

    @property
    def memory_request(self) -> str:
        return self._memory_request

    @memory_request.setter
    def memory_request(self, memory_request):
        self._memory_request = memory_request

    @property
    def gpu_request(self) -> Optional[str]:
        return self._gpu_request

    @gpu_request.setter
    def gpu_request(self, gpu_request):
        self._gpu_request = gpu_request

    @property
    def gpu_name(self) -> Optional[str]:
        return self._gpu_name

    @gpu_name.setter
    def gpu_name(self, gpu_name):
        self._gpu_name = gpu_name

    def validate(self):
        if self._min_replica > self._max_replica:
            raise Exception("Min replica must be less or equal to max replica")



# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from time import sleep
from typing import Union

import joblib
import merlin
import numpy as np
import pandas as pd
import pytest
from merlin.batch.config import PredictionJobConfig
from merlin.batch.job import JobStatus
from merlin.batch.sink import BigQuerySink, SaveMode
from merlin.batch.source import BigQuerySource
from merlin.model import ModelType, PyFuncV2Model
from sklearn import svm
from sklearn.datasets import load_iris

MODEL_PATH_ARTIFACT_KEY = "model_path"
MODEL_DIR = "test/batch/model"
MODEL_FILE = "model.joblib"
MODEL_PATH = os.path.join(MODEL_DIR, MODEL_FILE)
ENV_PATH = os.path.join(MODEL_DIR, "env.yaml")


class IrisClassifier(PyFuncV2Model):

    def initialize(self, artifacts: dict):
        self._model = joblib.load(artifacts[MODEL_PATH_ARTIFACT_KEY])

    def infer(self, model_input: pd.DataFrame) -> Union[np.ndarray,
                                                        pd.Series,
                                                        pd.DataFrame]:
        return self._model.predict(model_input)


@pytest.fixture
def batch_bigquery_source():
    return os.environ.get("E2E_BATCH_BIGQUERY_SOURCE", default="project.dataset.table")


@pytest.fixture
def batch_bigquery_sink():
    return os.environ.get("E2E_BATCH_BIGQUERY_SINK", default="project.dataset.table_result")


@pytest.fixture
def batch_gcs_staging_bucket():
    return os.environ.get("E2E_BATCH_GCS_STAGING_BUCKET", default="bucket-name")


@pytest.mark.batch
@pytest.mark.integration
def test_batch_pyfunc_v2_batch(integration_test_url, project_name, service_account, use_google_oauth,
                               batch_bigquery_source, batch_bigquery_sink, batch_gcs_staging_bucket):
    merlin.set_url(integration_test_url, use_google_oauth=use_google_oauth)
    merlin.set_project(project_name)
    merlin.set_model("batch-iris", ModelType.PYFUNC_V2)
    service_account_name = "merlin-integration-test@project.iam.gserviceaccount.com"
    _create_secret(merlin.active_project(),
                   service_account_name, service_account)

    clf = svm.SVC(gamma='scale')
    iris = load_iris()
    X, y = iris.data, iris.target
    clf.fit(X, y)
    joblib.dump(clf, MODEL_PATH)
    # Create new version of the model
    mdl = merlin.active_model()
    v = mdl.new_model_version()
    v.start()
    # Upload the serialized model to MLP
    v.log_pyfunc_model(model_instance=IrisClassifier(),
                       conda_env=ENV_PATH,
                       code_dir=["test"],
                       artifacts={MODEL_PATH_ARTIFACT_KEY: MODEL_PATH})

    v.finish()

    bq_source = BigQuerySource(batch_bigquery_source,
                               features=["sepal_length", "sepal_width", "petal_length", "petal_width"])
    bq_sink = BigQuerySink(batch_bigquery_sink,
                           staging_bucket=batch_gcs_staging_bucket,
                           result_column="prediction",
                           save_mode=SaveMode.OVERWRITE)
    job_config = PredictionJobConfig(
        source=bq_source, sink=bq_sink, service_account_name=service_account_name, env_vars={"ALPHA": "0.2"})
    job = v.create_prediction_job(job_config=job_config)

    assert job.status == JobStatus.COMPLETED

    job = v.create_prediction_job(job_config=job_config, sync=False)
    while job.status == JobStatus.PENDING:
        sleep(1)
        job = job.refresh()
    job = job.stop()

    assert job.status == JobStatus.TERMINATED


def _create_secret(project, secret_name, secret_value):
    secrets = project.list_secret()
    for secret in secrets:
        if secret == secret_name:
            return

    if secret_value is None:
        raise ValueError(
            f"{secret_name} secret is not found in the project and E2E_SERVICE_ACCOUNT is empty")
    project.create_secret(secret_name, secret_value)



# coding: utf-8

"""
    Merlin

    API Guide for accessing Merlin's model management, deployment, and serving functionalities  # noqa: E501

    OpenAPI spec version: 0.14.0
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

import pprint
import re  # noqa: F401

import six

class Config(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """
    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'job_config': 'PredictionJobConfig',
        'image_ref': 'str',
        'service_account_name': 'str',
        'resource_request': 'PredictionJobResourceRequest',
        'env_vars': 'list[EnvVar]'
    }

    attribute_map = {
        'job_config': 'job_config',
        'image_ref': 'image_ref',
        'service_account_name': 'service_account_name',
        'resource_request': 'resource_request',
        'env_vars': 'env_vars'
    }

    def __init__(self, job_config=None, image_ref=None, service_account_name=None, resource_request=None, env_vars=None):  # noqa: E501
        """Config - a model defined in Swagger"""  # noqa: E501
        self._job_config = None
        self._image_ref = None
        self._service_account_name = None
        self._resource_request = None
        self._env_vars = None
        self.discriminator = None
        if job_config is not None:
            self.job_config = job_config
        if image_ref is not None:
            self.image_ref = image_ref
        if service_account_name is not None:
            self.service_account_name = service_account_name
        if resource_request is not None:
            self.resource_request = resource_request
        if env_vars is not None:
            self.env_vars = env_vars

    @property
    def job_config(self):
        """Gets the job_config of this Config.  # noqa: E501


        :return: The job_config of this Config.  # noqa: E501
        :rtype: PredictionJobConfig
        """
        return self._job_config

    @job_config.setter
    def job_config(self, job_config):
        """Sets the job_config of this Config.


        :param job_config: The job_config of this Config.  # noqa: E501
        :type: PredictionJobConfig
        """

        self._job_config = job_config

    @property
    def image_ref(self):
        """Gets the image_ref of this Config.  # noqa: E501


        :return: The image_ref of this Config.  # noqa: E501
        :rtype: str
        """
        return self._image_ref

    @image_ref.setter
    def image_ref(self, image_ref):
        """Sets the image_ref of this Config.


        :param image_ref: The image_ref of this Config.  # noqa: E501
        :type: str
        """

        self._image_ref = image_ref

    @property
    def service_account_name(self):
        """Gets the service_account_name of this Config.  # noqa: E501


        :return: The service_account_name of this Config.  # noqa: E501
        :rtype: str
        """
        return self._service_account_name

    @service_account_name.setter
    def service_account_name(self, service_account_name):
        """Sets the service_account_name of this Config.


        :param service_account_name: The service_account_name of this Config.  # noqa: E501
        :type: str
        """

        self._service_account_name = service_account_name

    @property
    def resource_request(self):
        """Gets the resource_request of this Config.  # noqa: E501


        :return: The resource_request of this Config.  # noqa: E501
        :rtype: PredictionJobResourceRequest
        """
        return self._resource_request

    @resource_request.setter
    def resource_request(self, resource_request):
        """Sets the resource_request of this Config.


        :param resource_request: The resource_request of this Config.  # noqa: E501
        :type: PredictionJobResourceRequest
        """

        self._resource_request = resource_request

    @property
    def env_vars(self):
        """Gets the env_vars of this Config.  # noqa: E501


        :return: The env_vars of this Config.  # noqa: E501
        :rtype: list[EnvVar]
        """
        return self._env_vars

    @env_vars.setter
    def env_vars(self, env_vars):
        """Sets the env_vars of this Config.


        :param env_vars: The env_vars of this Config.  # noqa: E501
        :type: list[EnvVar]
        """

        self._env_vars = env_vars

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(Config, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, Config):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other



# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, Optional, Any

from merlin.protocol import Protocol
from merlin.resource_request import ResourceRequest
from merlin.util import autostr
from merlin import fluent

from enum import Enum

import client
import yaml
import json


class TransformerType(Enum):
    CUSTOM_TRANSFORMER = "custom"
    STANDARD_TRANSFORMER = "standard"


@autostr
class Transformer:
    StandardTransformerConfigKey = "STANDARD_TRANSFORMER_CONFIG"

    def __init__(
        self,
        image: str,
        enabled: bool = True,
        command: str = None,
        args: str = None,
        resource_request: ResourceRequest = None,
        env_vars: Dict[str, str] = None,
        transformer_type: TransformerType = TransformerType.CUSTOM_TRANSFORMER,
    ):
        self._image = image
        self._enabled = enabled
        self._command = command
        self._args = args
        self._resource_request = resource_request
        self._env_vars = env_vars
        self._transformer_type = transformer_type

    @property
    def image(self) -> str:
        return self._image

    @property
    def enabled(self) -> bool:
        return self._enabled

    @property
    def command(self) -> Optional[str]:
        return self._command

    @property
    def args(self) -> Optional[str]:
        return self._args

    @property
    def resource_request(self) -> Optional[ResourceRequest]:
        return self._resource_request

    @property
    def env_vars(self) -> Optional[Dict[str, str]]:
        return self._env_vars

    @property
    def transformer_type(self) -> TransformerType:
        return self._transformer_type


class StandardTransformer(Transformer):
    def __init__(
        self,
        config_file: str,
        enabled: bool = True,
        resource_request: ResourceRequest = None,
        env_vars: Dict[str, str] = None,
    ):
        self._load_transformer_config(config_file)
        transformer_env_var = self._transformer_config_env_var()
        merged_env_vars = env_vars or {}
        merged_env_vars = {**merged_env_vars, **transformer_env_var}
        super().__init__(
            image="",
            enabled=enabled,
            resource_request=resource_request,
            env_vars=merged_env_vars,
            transformer_type=TransformerType.STANDARD_TRANSFORMER,
        )

    def _load_transformer_config(self, config_file: str):
        with open(config_file, "r") as stream:
            self.transformer_config = yaml.safe_load(stream)

    def _transformer_config_env_var(self):
        config_json_string = json.dumps(self.transformer_config)
        return {self.StandardTransformerConfigKey: config_json_string}

    def simulate(
        self,
        payload: Dict,
        headers: Optional[Dict[Any, Any]] = None,
        model_prediction_config: Optional[Dict[Any, Any]] = None,
        protocol: str = "HTTP_JSON",
        exclude_tracing: bool = False,
    ) -> Dict:
        fluent._check_active_client()
        if not fluent._merlin_client:
            raise Exception("Merlin client is not initialized")

        response = fluent._merlin_client.standard_transformer_simulate(
            payload=payload,
            headers=headers,
            config=self.transformer_config,
            model_prediction_config=model_prediction_config,
            protocol=protocol,
        )

        # if exclude tracing delete key operation_tracing
        response = response.to_dict()
        if exclude_tracing:
            del response["operation_tracing"]

        return response



# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
import client
from merlin.util import autostr
from typing import Optional

class LoggerMode(Enum):
    ALL = 'all'
    REQUEST = 'request'
    RESPONSE = 'response'


@autostr
class LoggerConfig:
    def __init__(self, enabled: bool, mode: LoggerMode):
        self._enabled = enabled
        self._mode = mode

    @property
    def enabled(self):
        return self._enabled

    @property
    def mode(self):
        return self._mode
    
@autostr
class PredictionLoggerConfig:
    def __init__(self, enabled: bool, raw_features_table: str, entities_table: str) -> None:
        self._enabled = enabled
        self._raw_features_table = raw_features_table
        self._entities_table = entities_table

    @property
    def enabled(self):
        return self._enabled
    
    @property
    def raw_features_table(self):
        return self._raw_features_table
    
    @property
    def entities_table(self):
        return self._entities_table


@autostr
class Logger:
    logger_mode_mapping = {
        LoggerMode.ALL: client.LoggerMode.ALL,
        LoggerMode.REQUEST: client.LoggerMode.REQUEST,
        LoggerMode.RESPONSE: client.LoggerMode.RESPONSE
    }

    logger_mode_mapping_rev = {
        client.LoggerMode.ALL: LoggerMode.ALL,
        client.LoggerMode.REQUEST: LoggerMode.REQUEST,
        client.LoggerMode.RESPONSE: LoggerMode.RESPONSE,
    }

    def __init__(self, model: LoggerConfig = None, transformer: LoggerConfig = None, prediction: PredictionLoggerConfig = None):
        self._model = model
        self._transformer = transformer
        self._prediction = prediction

    @classmethod
    def from_logger_response(cls, response: client.Logger):
        if response is None:
            return Logger()
        model_config = None
        if response.model is not None:
            model_config = LoggerConfig(enabled=response.model.enabled, mode=cls._get_logger_mode_from_api_response(response.model.mode))
        transformer_config = None
        if response.transformer is not None:
            transformer_config = LoggerConfig(enabled=response.transformer.enabled,
                                              mode=cls._get_logger_mode_from_api_response(response.transformer.mode))
        prediction_config = None
        if response.prediction is not None:
            prediction_config = PredictionLoggerConfig(enabled=response.prediction.enabled,
                                                              raw_features_table=response.prediction.raw_features_table,
                                                              entities_table=response.prediction.entities_table)

        return Logger(model=model_config, transformer=transformer_config, prediction=prediction_config)

    @classmethod
    def _get_logger_mode_from_api_response(cls, mode_from_api_response):
        mode = cls.logger_mode_mapping_rev.get(mode_from_api_response)
        if mode is None:
            mode = LoggerMode.ALL
        return mode


    def to_logger_spec(self) -> Optional[client.Logger]:
        target_logger = None

        model_logger_config = None
        if self.model is not None:
            model_logger_config = client.LoggerConfig(
                enabled=self.model.enabled, mode=Logger.logger_mode_mapping[self.model.mode])

        transformer_logger_config = None
        if self.transformer is not None:
            transformer_logger_config = client.LoggerConfig(
                enabled=self.transformer.enabled, mode=Logger.logger_mode_mapping[self.transformer.mode])

        prediction_logger_config = None
        if self.prediction is not None:
            prediction_logger_config = client.PredictionLoggerConfig(
                enabled=self.prediction.enabled, raw_features_table= self.prediction.raw_features_table, entities_table=self.prediction.entities_table
            )

        if model_logger_config is not None or transformer_logger_config is not None or prediction_logger_config is not None:
            target_logger = client.Logger(model=model_logger_config, transformer=transformer_logger_config, prediction=prediction_logger_config)

        return target_logger


    @property
    def model(self):
        return self._model

    @property
    def transformer(self):
        return self._transformer
    
    @property
    def prediction(self):
        return self._prediction



# coding: utf-8

"""
    Merlin

    API Guide for accessing Merlin's model management, deployment, and serving functionalities  # noqa: E501

    OpenAPI spec version: 0.14.0
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

import pprint
import re  # noqa: F401

import six

class StandardTransformerSimulationResponse(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """
    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'response': 'FreeFormObject',
        'operation_tracing': 'OperationTracing'
    }

    attribute_map = {
        'response': 'response',
        'operation_tracing': 'operation_tracing'
    }

    def __init__(self, response=None, operation_tracing=None):  # noqa: E501
        """StandardTransformerSimulationResponse - a model defined in Swagger"""  # noqa: E501
        self._response = None
        self._operation_tracing = None
        self.discriminator = None
        if response is not None:
            self.response = response
        if operation_tracing is not None:
            self.operation_tracing = operation_tracing

    @property
    def response(self):
        """Gets the response of this StandardTransformerSimulationResponse.  # noqa: E501


        :return: The response of this StandardTransformerSimulationResponse.  # noqa: E501
        :rtype: FreeFormObject
        """
        return self._response

    @response.setter
    def response(self, response):
        """Sets the response of this StandardTransformerSimulationResponse.


        :param response: The response of this StandardTransformerSimulationResponse.  # noqa: E501
        :type: FreeFormObject
        """

        self._response = response

    @property
    def operation_tracing(self):
        """Gets the operation_tracing of this StandardTransformerSimulationResponse.  # noqa: E501


        :return: The operation_tracing of this StandardTransformerSimulationResponse.  # noqa: E501
        :rtype: OperationTracing
        """
        return self._operation_tracing

    @operation_tracing.setter
    def operation_tracing(self, operation_tracing):
        """Sets the operation_tracing of this StandardTransformerSimulationResponse.


        :param operation_tracing: The operation_tracing of this StandardTransformerSimulationResponse.  # noqa: E501
        :type: OperationTracing
        """

        self._operation_tracing = operation_tracing

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(StandardTransformerSimulationResponse, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, StandardTransformerSimulationResponse):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other



# coding: utf-8

"""
    Merlin

    API Guide for accessing Merlin's model management, deployment, and serving functionalities  # noqa: E501

    OpenAPI spec version: 0.14.0
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

import pprint
import re  # noqa: F401

import six

class PredictionJobConfigGcsSource(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """
    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'format': 'FileFormat',
        'uri': 'str',
        'features': 'list[str]',
        'options': 'dict(str, str)'
    }

    attribute_map = {
        'format': 'format',
        'uri': 'uri',
        'features': 'features',
        'options': 'options'
    }

    def __init__(self, format=None, uri=None, features=None, options=None):  # noqa: E501
        """PredictionJobConfigGcsSource - a model defined in Swagger"""  # noqa: E501
        self._format = None
        self._uri = None
        self._features = None
        self._options = None
        self.discriminator = None
        if format is not None:
            self.format = format
        if uri is not None:
            self.uri = uri
        if features is not None:
            self.features = features
        if options is not None:
            self.options = options

    @property
    def format(self):
        """Gets the format of this PredictionJobConfigGcsSource.  # noqa: E501


        :return: The format of this PredictionJobConfigGcsSource.  # noqa: E501
        :rtype: FileFormat
        """
        return self._format

    @format.setter
    def format(self, format):
        """Sets the format of this PredictionJobConfigGcsSource.


        :param format: The format of this PredictionJobConfigGcsSource.  # noqa: E501
        :type: FileFormat
        """

        self._format = format

    @property
    def uri(self):
        """Gets the uri of this PredictionJobConfigGcsSource.  # noqa: E501


        :return: The uri of this PredictionJobConfigGcsSource.  # noqa: E501
        :rtype: str
        """
        return self._uri

    @uri.setter
    def uri(self, uri):
        """Sets the uri of this PredictionJobConfigGcsSource.


        :param uri: The uri of this PredictionJobConfigGcsSource.  # noqa: E501
        :type: str
        """

        self._uri = uri

    @property
    def features(self):
        """Gets the features of this PredictionJobConfigGcsSource.  # noqa: E501


        :return: The features of this PredictionJobConfigGcsSource.  # noqa: E501
        :rtype: list[str]
        """
        return self._features

    @features.setter
    def features(self, features):
        """Sets the features of this PredictionJobConfigGcsSource.


        :param features: The features of this PredictionJobConfigGcsSource.  # noqa: E501
        :type: list[str]
        """

        self._features = features

    @property
    def options(self):
        """Gets the options of this PredictionJobConfigGcsSource.  # noqa: E501


        :return: The options of this PredictionJobConfigGcsSource.  # noqa: E501
        :rtype: dict(str, str)
        """
        return self._options

    @options.setter
    def options(self, options):
        """Sets the options of this PredictionJobConfigGcsSource.


        :param options: The options of this PredictionJobConfigGcsSource.  # noqa: E501
        :type: dict(str, str)
        """

        self._options = options

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(PredictionJobConfigGcsSource, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, PredictionJobConfigGcsSource):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other



# coding: utf-8

"""
    Merlin

    API Guide for accessing Merlin's model management, deployment, and serving functionalities  # noqa: E501

    OpenAPI spec version: 0.14.0
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

import pprint
import re  # noqa: F401

import six

class VersionEndpoint(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """
    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'id': 'str',
        'version_id': 'int',
        'status': 'EndpointStatus',
        'url': 'str',
        'service_name': 'str',
        'environment_name': 'str',
        'environment': 'Environment',
        'monitoring_url': 'str',
        'message': 'str',
        'resource_request': 'ResourceRequest',
        'env_vars': 'list[EnvVar]',
        'transformer': 'Transformer',
        'logger': 'Logger',
        'deployment_mode': 'DeploymentMode',
        'autoscaling_policy': 'AutoscalingPolicy',
        'protocol': 'Protocol',
        'created_at': 'datetime',
        'updated_at': 'datetime'
    }

    attribute_map = {
        'id': 'id',
        'version_id': 'version_id',
        'status': 'status',
        'url': 'url',
        'service_name': 'service_name',
        'environment_name': 'environment_name',
        'environment': 'environment',
        'monitoring_url': 'monitoring_url',
        'message': 'message',
        'resource_request': 'resource_request',
        'env_vars': 'env_vars',
        'transformer': 'transformer',
        'logger': 'logger',
        'deployment_mode': 'deployment_mode',
        'autoscaling_policy': 'autoscaling_policy',
        'protocol': 'protocol',
        'created_at': 'created_at',
        'updated_at': 'updated_at'
    }

    def __init__(self, id=None, version_id=None, status=None, url=None, service_name=None, environment_name=None, environment=None, monitoring_url=None, message=None, resource_request=None, env_vars=None, transformer=None, logger=None, deployment_mode=None, autoscaling_policy=None, protocol=None, created_at=None, updated_at=None):  # noqa: E501
        """VersionEndpoint - a model defined in Swagger"""  # noqa: E501
        self._id = None
        self._version_id = None
        self._status = None
        self._url = None
        self._service_name = None
        self._environment_name = None
        self._environment = None
        self._monitoring_url = None
        self._message = None
        self._resource_request = None
        self._env_vars = None
        self._transformer = None
        self._logger = None
        self._deployment_mode = None
        self._autoscaling_policy = None
        self._protocol = None
        self._created_at = None
        self._updated_at = None
        self.discriminator = None
        if id is not None:
            self.id = id
        if version_id is not None:
            self.version_id = version_id
        if status is not None:
            self.status = status
        if url is not None:
            self.url = url
        if service_name is not None:
            self.service_name = service_name
        if environment_name is not None:
            self.environment_name = environment_name
        if environment is not None:
            self.environment = environment
        if monitoring_url is not None:
            self.monitoring_url = monitoring_url
        if message is not None:
            self.message = message
        if resource_request is not None:
            self.resource_request = resource_request
        if env_vars is not None:
            self.env_vars = env_vars
        if transformer is not None:
            self.transformer = transformer
        if logger is not None:
            self.logger = logger
        if deployment_mode is not None:
            self.deployment_mode = deployment_mode
        if autoscaling_policy is not None:
            self.autoscaling_policy = autoscaling_policy
        if protocol is not None:
            self.protocol = protocol
        if created_at is not None:
            self.created_at = created_at
        if updated_at is not None:
            self.updated_at = updated_at

    @property
    def id(self):
        """Gets the id of this VersionEndpoint.  # noqa: E501


        :return: The id of this VersionEndpoint.  # noqa: E501
        :rtype: str
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this VersionEndpoint.


        :param id: The id of this VersionEndpoint.  # noqa: E501
        :type: str
        """

        self._id = id

    @property
    def version_id(self):
        """Gets the version_id of this VersionEndpoint.  # noqa: E501


        :return: The version_id of this VersionEndpoint.  # noqa: E501
        :rtype: int
        """
        return self._version_id

    @version_id.setter
    def version_id(self, version_id):
        """Sets the version_id of this VersionEndpoint.


        :param version_id: The version_id of this VersionEndpoint.  # noqa: E501
        :type: int
        """

        self._version_id = version_id

    @property
    def status(self):
        """Gets the status of this VersionEndpoint.  # noqa: E501


        :return: The status of this VersionEndpoint.  # noqa: E501
        :rtype: EndpointStatus
        """
        return self._status

    @status.setter
    def status(self, status):
        """Sets the status of this VersionEndpoint.


        :param status: The status of this VersionEndpoint.  # noqa: E501
        :type: EndpointStatus
        """

        self._status = status

    @property
    def url(self):
        """Gets the url of this VersionEndpoint.  # noqa: E501


        :return: The url of this VersionEndpoint.  # noqa: E501
        :rtype: str
        """
        return self._url

    @url.setter
    def url(self, url):
        """Sets the url of this VersionEndpoint.


        :param url: The url of this VersionEndpoint.  # noqa: E501
        :type: str
        """

        self._url = url

    @property
    def service_name(self):
        """Gets the service_name of this VersionEndpoint.  # noqa: E501


        :return: The service_name of this VersionEndpoint.  # noqa: E501
        :rtype: str
        """
        return self._service_name

    @service_name.setter
    def service_name(self, service_name):
        """Sets the service_name of this VersionEndpoint.


        :param service_name: The service_name of this VersionEndpoint.  # noqa: E501
        :type: str
        """

        self._service_name = service_name

    @property
    def environment_name(self):
        """Gets the environment_name of this VersionEndpoint.  # noqa: E501


        :return: The environment_name of this VersionEndpoint.  # noqa: E501
        :rtype: str
        """
        return self._environment_name

    @environment_name.setter
    def environment_name(self, environment_name):
        """Sets the environment_name of this VersionEndpoint.


        :param environment_name: The environment_name of this VersionEndpoint.  # noqa: E501
        :type: str
        """

        self._environment_name = environment_name

    @property
    def environment(self):
        """Gets the environment of this VersionEndpoint.  # noqa: E501


        :return: The environment of this VersionEndpoint.  # noqa: E501
        :rtype: Environment
        """
        return self._environment

    @environment.setter
    def environment(self, environment):
        """Sets the environment of this VersionEndpoint.


        :param environment: The environment of this VersionEndpoint.  # noqa: E501
        :type: Environment
        """

        self._environment = environment

    @property
    def monitoring_url(self):
        """Gets the monitoring_url of this VersionEndpoint.  # noqa: E501


        :return: The monitoring_url of this VersionEndpoint.  # noqa: E501
        :rtype: str
        """
        return self._monitoring_url

    @monitoring_url.setter
    def monitoring_url(self, monitoring_url):
        """Sets the monitoring_url of this VersionEndpoint.


        :param monitoring_url: The monitoring_url of this VersionEndpoint.  # noqa: E501
        :type: str
        """

        self._monitoring_url = monitoring_url

    @property
    def message(self):
        """Gets the message of this VersionEndpoint.  # noqa: E501


        :return: The message of this VersionEndpoint.  # noqa: E501
        :rtype: str
        """
        return self._message

    @message.setter
    def message(self, message):
        """Sets the message of this VersionEndpoint.


        :param message: The message of this VersionEndpoint.  # noqa: E501
        :type: str
        """

        self._message = message

    @property
    def resource_request(self):
        """Gets the resource_request of this VersionEndpoint.  # noqa: E501


        :return: The resource_request of this VersionEndpoint.  # noqa: E501
        :rtype: ResourceRequest
        """
        return self._resource_request

    @resource_request.setter
    def resource_request(self, resource_request):
        """Sets the resource_request of this VersionEndpoint.


        :param resource_request: The resource_request of this VersionEndpoint.  # noqa: E501
        :type: ResourceRequest
        """

        self._resource_request = resource_request

    @property
    def env_vars(self):
        """Gets the env_vars of this VersionEndpoint.  # noqa: E501


        :return: The env_vars of this VersionEndpoint.  # noqa: E501
        :rtype: list[EnvVar]
        """
        return self._env_vars

    @env_vars.setter
    def env_vars(self, env_vars):
        """Sets the env_vars of this VersionEndpoint.


        :param env_vars: The env_vars of this VersionEndpoint.  # noqa: E501
        :type: list[EnvVar]
        """

        self._env_vars = env_vars

    @property
    def transformer(self):
        """Gets the transformer of this VersionEndpoint.  # noqa: E501


        :return: The transformer of this VersionEndpoint.  # noqa: E501
        :rtype: Transformer
        """
        return self._transformer

    @transformer.setter
    def transformer(self, transformer):
        """Sets the transformer of this VersionEndpoint.


        :param transformer: The transformer of this VersionEndpoint.  # noqa: E501
        :type: Transformer
        """

        self._transformer = transformer

    @property
    def logger(self):
        """Gets the logger of this VersionEndpoint.  # noqa: E501


        :return: The logger of this VersionEndpoint.  # noqa: E501
        :rtype: Logger
        """
        return self._logger

    @logger.setter
    def logger(self, logger):
        """Sets the logger of this VersionEndpoint.


        :param logger: The logger of this VersionEndpoint.  # noqa: E501
        :type: Logger
        """

        self._logger = logger

    @property
    def deployment_mode(self):
        """Gets the deployment_mode of this VersionEndpoint.  # noqa: E501


        :return: The deployment_mode of this VersionEndpoint.  # noqa: E501
        :rtype: DeploymentMode
        """
        return self._deployment_mode

    @deployment_mode.setter
    def deployment_mode(self, deployment_mode):
        """Sets the deployment_mode of this VersionEndpoint.


        :param deployment_mode: The deployment_mode of this VersionEndpoint.  # noqa: E501
        :type: DeploymentMode
        """

        self._deployment_mode = deployment_mode

    @property
    def autoscaling_policy(self):
        """Gets the autoscaling_policy of this VersionEndpoint.  # noqa: E501


        :return: The autoscaling_policy of this VersionEndpoint.  # noqa: E501
        :rtype: AutoscalingPolicy
        """
        return self._autoscaling_policy

    @autoscaling_policy.setter
    def autoscaling_policy(self, autoscaling_policy):
        """Sets the autoscaling_policy of this VersionEndpoint.


        :param autoscaling_policy: The autoscaling_policy of this VersionEndpoint.  # noqa: E501
        :type: AutoscalingPolicy
        """

        self._autoscaling_policy = autoscaling_policy

    @property
    def protocol(self):
        """Gets the protocol of this VersionEndpoint.  # noqa: E501


        :return: The protocol of this VersionEndpoint.  # noqa: E501
        :rtype: Protocol
        """
        return self._protocol

    @protocol.setter
    def protocol(self, protocol):
        """Sets the protocol of this VersionEndpoint.


        :param protocol: The protocol of this VersionEndpoint.  # noqa: E501
        :type: Protocol
        """

        self._protocol = protocol

    @property
    def created_at(self):
        """Gets the created_at of this VersionEndpoint.  # noqa: E501


        :return: The created_at of this VersionEndpoint.  # noqa: E501
        :rtype: datetime
        """
        return self._created_at

    @created_at.setter
    def created_at(self, created_at):
        """Sets the created_at of this VersionEndpoint.


        :param created_at: The created_at of this VersionEndpoint.  # noqa: E501
        :type: datetime
        """

        self._created_at = created_at

    @property
    def updated_at(self):
        """Gets the updated_at of this VersionEndpoint.  # noqa: E501


        :return: The updated_at of this VersionEndpoint.  # noqa: E501
        :rtype: datetime
        """
        return self._updated_at

    @updated_at.setter
    def updated_at(self, updated_at):
        """Sets the updated_at of this VersionEndpoint.


        :param updated_at: The updated_at of this VersionEndpoint.  # noqa: E501
        :type: datetime
        """

        self._updated_at = updated_at

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(VersionEndpoint, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, VersionEndpoint):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other



# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import json
from sys import version_info
from unittest import mock

import pytest

import client as cl
from client import ApiClient, Configuration
from merlin.client import MerlinClient
from merlin.endpoint import Status
from merlin.model import Model, ModelType, Project
from merlin.util import guess_mlp_ui_url
from merlin.version import VERSION

# get global mock responses that configured in conftest
responses = pytest.responses


@pytest.fixture
def mock_url():
    return "http://merlin.dev/api"


@pytest.fixture
def api_client(mock_url):
    config = Configuration()
    config.host = mock_url + "/v1"
    return ApiClient(config)


mlflow_tracking_url = "http://mlflow.api.merlin.dev"
created_at = "2019-08-29T08:13:12.377Z"
updated_at = "2019-08-29T08:13:12.377Z"

default_resource_request = cl.ResourceRequest(1, 1, "100m", "128Mi")
env_1 = cl.Environment(
    id=1,
    name="dev",
    cluster="cluster-1",
    is_default=True,
    default_resource_request=default_resource_request,
)
env_2 = cl.Environment(
    id=2,
    name="dev-2",
    cluster="cluster-2",
    is_default=False,
    default_resource_request=default_resource_request,
)
mdl_endpoint_1 = cl.ModelEndpoint(
    id=1,
    model_id=1,
    model=None,
    status="serving",
    url="localhost",
    rule=None,
    environment_name=env_1.name,
    environment=env_1,
    created_at=created_at,
    updated_at=updated_at,
)


@responses.activate
def test_get_project(mock_url, mock_oauth, use_google_oauth):
    responses.add(
        "GET",
        "/api/v1/projects",
        body=f"""[{{
                        "id": 0,
                        "name": "my-project",
                        "mlflow_tracking_url": "http://mlflow.api.merlin.dev",
                        "created_at": "{created_at}",
                        "updated_at": "{updated_at}"
                      }}]""",
        status=200,
        content_type="application/json",
    )

    m = MerlinClient(mock_url, use_google_oauth=use_google_oauth)
    p = m.get_project("my-project")

    assert responses.calls[-1].request.method == "GET"
    assert responses.calls[-1].request.url == "/api/v1/projects?name=my-project"
    assert responses.calls[-1].request.host == "merlin.dev"

    assert p.id == 0
    assert p.name == "my-project"
    assert p.mlflow_tracking_url == "http://mlflow.api.merlin.dev"
    assert p.url == mock_url
    assert isinstance(p.created_at, datetime.datetime)
    assert isinstance(p.updated_at, datetime.datetime)


@responses.activate
def test_create_invalid_project_name(
    mock_url, api_client, mock_oauth, use_google_oauth
):
    project_name = "invalidProjectName"

    client = MerlinClient(mock_url, use_google_oauth=use_google_oauth)

    # Try to create project with invalid name. It must be fail
    with pytest.raises(Exception):
        assert client.get_project(project_name)


@responses.activate
def test_create_model(mock_url, api_client, mock_oauth, use_google_oauth):
    project_id = "1010"
    mlflow_experiment_id = 1
    model_name = "my-model"
    project_name = "my-project"
    model_type = ModelType.XGBOOST
    mlflow_url = "http://mlflow.api.merlin.dev"

    responses.add(
        "GET",
        f"/api/v1/projects/{project_id}/models",
        body="[]",
        status=200,
        content_type="application/json",
    )
    responses.add(
        "POST",
        f"/api/v1/projects/{project_id}/models",
        body=f"""{{
                        "id": 0,
                        "project_id": {project_id},
                        "mlflow_experiment_id": {mlflow_experiment_id},
                        "name": "{model_name}",
                        "type": "{model_type.value}",
                        "mlflow_url": "{mlflow_url}",
                        "endpoints": [],
                        "created_at": "{created_at}",
                        "updated_at": "{updated_at}"
                      }}""",
        status=200,
        content_type="application/json",
    )

    client = MerlinClient(mock_url, use_google_oauth=use_google_oauth)
    prj = cl.Project(
        project_id, project_name, mlflow_tracking_url, created_at, updated_at
    )
    project = Project(prj, mock_url, api_client)
    with mock.patch.object(client, "get_project", return_value=project):
        model = client.get_or_create_model(
            "my-model", project_name=project_name, model_type=model_type
        )

        assert json.loads(responses.calls[-1].request.body) == json.loads(
            f"""
        {{
            "name" : "{model_name}",
            "type" : "{model_type.value}"
        }}
        """
        )
        assert model.id == 0
        assert model.mlflow_experiment_id == mlflow_experiment_id
        assert model.name == model_name
        assert model.type == model_type
        assert model.mlflow_url == mlflow_tracking_url
        assert model.mlflow_experiment_id == mlflow_experiment_id
        assert isinstance(model.created_at, datetime.datetime)
        assert isinstance(model.updated_at, datetime.datetime)
        assert model.project == project
        assert (
            f"merlin-sdk/{VERSION}" in responses.calls[-1].request.headers["User-Agent"]
        )
        assert (
            f"python/{version_info.major}.{version_info.minor}.{version_info.micro}"
            in responses.calls[-1].request.headers["User-Agent"]
        )


@responses.activate
def test_create_invalid_model_name(mock_url, api_client, mock_oauth, use_google_oauth):
    model_name = "invalidModelName"
    project_name = "my-project"
    model_type = ModelType.XGBOOST

    client = MerlinClient(mock_url, use_google_oauth=use_google_oauth)

    # Try to create model with invalid name. It must be fail
    with pytest.raises(Exception):
        assert client.get_or_create_model(model_name, project_name, model_type)


@responses.activate
def test_get_model(mock_url, api_client, mock_oauth, use_google_oauth):
    project_id = "1010"
    mlflow_experiment_id = 1
    model_name = "my-model"
    project_name = "my-project"
    model_type = ModelType.XGBOOST
    mlflow_url = "http://mlflow.api.merlin.dev"

    responses.add(
        "GET",
        f"/api/v1/projects/{project_id}/models",
        body=f"""[{{
                        "id": 1,
                        "project_id": {project_id},
                        "mlflow_experiment_id": {mlflow_experiment_id},
                        "name": "{model_name}",
                        "type": "{model_type.value}",
                        "mlflow_url": "{mlflow_url}",
                        "endpoints": [],
                        "created_at": "{created_at}",
                        "updated_at": "{updated_at}"
                      }}]""",
        status=200,
        content_type="application/json",
    )

    responses.add(
        "GET",
        f"/api/v1/models/1/endpoints",
        body=json.dumps([mdl_endpoint_1.to_dict()]),
        status=200,
        content_type="application/json",
    )

    client = MerlinClient(mock_url, use_google_oauth=use_google_oauth)
    prj = cl.Project(
        project_id, project_name, mlflow_tracking_url, created_at, updated_at
    )
    project = Project(prj, mock_url, api_client)
    with mock.patch.object(client, "get_project", return_value=project):
        model = client.get_or_create_model(
            "my-model", project_name=project_name, model_type=model_type
        )
        assert model.id == 1
        assert model.name == model_name
        assert model.type == model_type
        assert model.mlflow_url == mlflow_tracking_url
        assert model.mlflow_experiment_id == mlflow_experiment_id
        assert isinstance(model.created_at, datetime.datetime)
        assert isinstance(model.updated_at, datetime.datetime)
        assert model.project == project

        default_model_endpoint = model.endpoint
        assert default_model_endpoint is not None
        assert default_model_endpoint.status == Status.SERVING
        assert default_model_endpoint.environment_name == env_1.name


@responses.activate
def test_new_model_version(mock_url, api_client, mock_oauth, use_google_oauth):
    project_id = 1
    model_id = 1
    version_id = 2
    model_name = "my-model"
    project_name = "my-project"
    mlflow_experiment_id = 1
    mlflow_run_id = "c5c3b6b220b34c7496de8c0400b7c793"
    model_type = ModelType.TENSORFLOW
    mlflow_url = "http://mlflow.api.merlin.dev"
    artifact_uri = "gs://zltest/model"
    created_at = "2019-09-04T03:09:13.842Z"
    updated_at = "2019-09-04T03:09:13.843Z"

    responses.add(
        "POST",
        f"/api/v1/models/{model_id}/versions",
        body=f"""{{
                        "id": {version_id},
                        "model_id": {model_id},
                        "mlflow_run_id": "{mlflow_run_id}",
                        "mlflow_url": "{mlflow_url}",
                        "artifact_uri": "{artifact_uri}",
                        "endpoints": [],
                        "mlflow_url": "{mlflow_url}",
                        "created_at": "{created_at}",
                        "updated_at": "{updated_at}"
                      }}""",
        status=200,
        content_type="application/json",
    )

    client = MerlinClient(mock_url, use_google_oauth=use_google_oauth)
    prj = cl.Project(
        project_id, project_name, mlflow_tracking_url, created_at, updated_at
    )
    project = Project(prj, mock_url, api_client)
    mdl = cl.Model(
        model_id,
        project_id,
        mlflow_experiment_id,
        model_name,
        model_type.value,
        mlflow_url,
        None,
        created_at,
        updated_at,
    )
    mdl = Model(mdl, project, api_client)
    with mock.patch.object(client, "get_model", return_value=mdl):
        mv = client.new_model_version(model_name, project_name)

        assert mv.id == version_id
        assert mv.mlflow_run_id == mlflow_run_id
        assert mv.mlflow_url == mlflow_url
        assert mv.properties is None
        assert isinstance(mv.created_at, datetime.datetime)
        assert isinstance(mv.updated_at, datetime.datetime)
        assert mv.model == mdl
        ui_url = guess_mlp_ui_url(mock_url)
        assert mv.url == f"{ui_url}/projects/1/models/{model_id}/versions"


@responses.activate
def test_list_environments(mock_url, api_client, mock_oauth, use_google_oauth):
    responses.add(
        "GET",
        "/api/v1/environments",
        body=json.dumps([env_1.to_dict(), env_2.to_dict()]),
        status=200,
        content_type="application/json",
    )
    client = MerlinClient(mock_url, use_google_oauth=use_google_oauth)
    envs = client.list_environment()

    assert len(envs) == 2
    assert envs[0].name == env_1.name
    assert envs[0].cluster == env_1.cluster
    assert envs[0].is_default == env_1.is_default
    assert envs[1].name == env_2.name
    assert envs[1].cluster == env_2.cluster
    assert envs[1].is_default == env_2.is_default


@responses.activate
def test_get_environment(mock_url, api_client, mock_oauth, use_google_oauth):
    responses.add(
        "GET",
        "/api/v1/environments",
        body=json.dumps([env_1.to_dict(), env_2.to_dict()]),
        status=200,
        content_type="application/json",
    )
    client = MerlinClient(mock_url, use_google_oauth=use_google_oauth)
    env = client.get_environment(env_1.name)

    assert env is not None
    assert env.name == env_1.name
    assert env.cluster == env_1.cluster
    assert env.is_default == env_1.is_default

    env = client.get_environment("undefined_env")
    assert env is None


@responses.activate
def test_get_default_environment(mock_url, api_client, mock_oauth, use_google_oauth):
    responses.add(
        "GET",
        "/api/v1/environments",
        body=json.dumps([env_1.to_dict(), env_2.to_dict()]),
        status=200,
        content_type="application/json",
    )
    client = MerlinClient(mock_url, use_google_oauth=use_google_oauth)
    env = client.get_default_environment()

    assert env.name == env_1.name
    assert env.cluster == env_1.cluster
    assert env.is_default == env_1.is_default

    responses.reset()

    responses.add(
        "GET",
        "/api/v1/environments",
        body=json.dumps([env_2.to_dict()]),
        status=200,
        content_type="application/json",
    )
    env = client.get_default_environment()

    assert env is None


@responses.activate
def test_get_default_environment(mock_url, api_client, mock_oauth, use_google_oauth):
    client = MerlinClient(mock_url, use_google_oauth=use_google_oauth)
    responses.add(
        "GET",
        "/api/v1/environments",
        body=json.dumps([env_2.to_dict()]),
        status=200,
        content_type="application/json",
    )
    env = client.get_default_environment()

    assert env is None



from abc import abstractmethod
from typing import Union

import grpc
import numpy
import pandas
from caraml.upi.v1 import upi_pb2
from mlflow.pyfunc import PythonModel

from merlin.protocol import Protocol

PYFUNC_EXTRA_ARGS_KEY = "__EXTRA_ARGS__"
PYFUNC_MODEL_INPUT_KEY = "__INPUT__"
PYFUNC_PROTOCOL_KEY = "__PROTOCOL__"
PYFUNC_GRPC_CONTEXT = "__GRPC_CONTEXT__"


class PyFuncModel(PythonModel):
    def load_context(self, context):
        """
        Override method of PythonModel `load_context`. This method load artifacts from context
        that can be used in predict function. This method is called by internal MLflow when an MLflow
        is loaded.

        :param context: A :class:`~PythonModelContext` instance containing artifacts that the model
                        can use to perform inference
        """
        self.initialize(context.artifacts)
        self._use_kwargs_infer = True

    def predict(self, context, model_input):
        """
        Implementation of PythonModel `predict` method. This method evaluates model input and produces model output.

        :param context: A :class:`~PythonModelContext` instance containing artifacts that the model
                        can use to perform inference
        :param model_input: A pyfunc-compatible input for the model to evaluate.
        """
        extra_args = model_input.get(PYFUNC_EXTRA_ARGS_KEY, {})
        input = model_input.get(PYFUNC_MODEL_INPUT_KEY, {})
        protocol = model_input.get(PYFUNC_PROTOCOL_KEY, Protocol.HTTP_JSON)

        if protocol == Protocol.HTTP_JSON:
            if extra_args is not None:
                return self._do_http_predict(input, **extra_args)
            return self._do_http_predict(input)
        elif protocol == Protocol.UPI_V1:
            grpc_context = model_input.get(PYFUNC_GRPC_CONTEXT, None)
            return self.upiv1_infer(input, grpc_context)
        else:
            raise NotImplementedError(f"protocol {protocol} is not supported")

    def _do_http_predict(self, model_input, **kwargs):
        if self._use_kwargs_infer:
            try:
                return self.infer(model_input, **kwargs)
            except TypeError as e:
                if "infer() got an unexpected keyword argument" in str(e):
                    print(
                        'Fallback to the old infer() method, got TypeError exception: {}'.format(e))
                    self._use_kwargs_infer = False
                else:
                    raise e

        return self.infer(model_input)

    @abstractmethod
    def initialize(self, artifacts: dict):
        """
        Implementation of PyFuncModel can specify initialization step which
        will be called one time during model initialization.

        :param artifacts: dictionary of artifacts passed to log_model method
        """
        pass

    @abstractmethod
    def infer(self, request: dict, **kwargs) -> dict:
        """
        Do inference.

        This method MUST be implemented by concrete implementation of
        PyFuncModel to support HTTP_JSON protocol.
        This method accept 'request' which is the body content of incoming
        request.
        Implementation should return inference a json object of response.

        :param request: Dictionary containing incoming request body content
        :param **kwargs: See below.

        :return: Dictionary containing response body

        :keyword arguments:
        * headers (dict): Dictionary containing incoming HTTP request headers
        """
        raise NotImplementedError("infer is not implemented")

    @abstractmethod
    def upiv1_infer(self, request: upi_pb2.PredictValuesRequest,
                    context: grpc.ServicerContext) -> upi_pb2.PredictValuesResponse:
        """
        Do inference.

        This method MUST be implemented by concrete implementation of PyFunModel to support UPI_V1 protocol.
        The method accept request in for of PredictValuesRequest proto and should return PredictValuesResponse response proto.

        :param request: Inference request as PredictValuesRequest
        :param context: grpc context
        :return: Prediction result as PredictValuesResponse proto
        """
        raise NotImplementedError("upiv1_infer is not implemented")



class PyFuncV2Model(PythonModel):
    def load_context(self, context):
        """
        Override method of PythonModel `load_context`. This method load artifacts from context
        that can be used in predict function. This method is called by internal MLflow when an MLflow
        is loaded.

        :param context: A :class:`~PythonModelContext` instance containing artifacts that the model
                        can use to perform inference
        """
        self.initialize(context.artifacts)

    def predict(self, context, model_input):
        """
        Implementation of PythonModel `predict` method. This method evaluates model input and produces model output.

        :param context: A :class:`~PythonModelContext` instance containing artifacts that the model
                        can use to perform inference
        :param model_input: A pyfunc-compatible input for the model to evaluate.
        """
        return self.infer(model_input)

    def initialize(self, artifacts: dict):
        """
        Implementation of PyFuncModel can specify initialization step which
        will be called one time during model initialization.

        :param artifacts: dictionary of artifacts passed to log_model method
        """
        pass

    def infer(self, model_input: pandas.DataFrame) -> Union[numpy.ndarray,
                                                            pandas.Series,
                                                            pandas.DataFrame]:
        """
        Infer method is the main method that will be called when calculating
        the inference result for both online prediction and batch
        prediction. The method accepts pandas Dataframe and returns either
        another panda Dataframe / pandas Series / ndarray of the same length
        as the input. In the batch prediction case the model_input will
        contain an arbitrary partition of the whole dataset that the user
        defines as the data source. As such, it is advisable not to do
        aggregation within the infer method, as it will be incorrect since
        it will only apply to the partition in contrary to the whole dataset.

        :param model_input: input to the model (pandas.DataFrame)
        :return: inference result as numpy.ndarray or pandas.Series or pandas.DataFrame

        """
        raise NotImplementedError("infer is not implemented")

    def preprocess(self, request: dict) -> pandas.DataFrame:
        """
        Preprocess incoming request into a pandas Dataframe that will be
        passed to the infer method.
        This method will not be called during batch prediction.

        :param request: dictionary representing the incoming request body
        :return: pandas.DataFrame that will be passed to infer method
        """
        raise NotImplementedError("preprocess is not implemented")

    def postprocess(self, model_result: Union[numpy.ndarray,
                                              pandas.Series,
                                              pandas.DataFrame]) -> dict:
        """
        Postprocess prediction result returned by infer method into
        dictionary representing the response body of the model.
        This method will not be called during batch prediction.

        :param model_result: output of the model's infer method
        :return: dictionary containing the response body
        """
        raise NotImplementedError("postprocess is not implemented")

    def raw_infer(self, request: dict) -> dict:
        """
        Do inference
        This method MUST be implemented by concrete implementation of
        PyFuncV2Model.
        This method accept 'request' which is the body content of incoming
        request.
        This method will not be called during batch prediction.

        Implementation should return inference a json object of response.

        :param request: Dictionary containing incoming request body content
        :return: Dictionary containing response body
        """
        raise NotImplementedError("raw_infer is not implemented")



#!/usr/bin/env python
# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import imp
import os

from setuptools import find_packages, setup

version = imp.load_source(
    'merlin.version', os.path.join('merlin', 'version.py')).VERSION

REQUIRES = [
    "boto3>=1.9.84",
    "caraml-upi-protos>=0.3.1",
    "certifi>=2017.4.17",
    "Click>=7.0,<8.1.4",
    "cloudpickle==2.0.0",  # used by mlflow
    "cookiecutter>=1.7.2",
    "docker>=4.2.1",
    "google-cloud-storage>=1.19.0",
    "protobuf>=3.12.0,<5.0.0", # Determined by the mlflow dependency
    "mlflow>=1.26.1,<2.0.0",
    "PyPrind>=2.11.2",
    "python_dateutil>=2.5.3",
    "PyYAML>=5.4",
    "six>=1.10",
    "urllib3>=1.26",
    "numpy<=1.23.5", # Temporary pin numpy due to https://numpy.org/doc/stable/release/1.20.0-notes.html#numpy-1-20-0-release-notes
    "caraml-auth-google==0.0.0.post7",
]

TEST_REQUIRES = [
    "google-cloud-bigquery-storage>=0.7.0",
    "google-cloud-bigquery>=1.18.0",
    "joblib>=0.13.0,<1.2.0",  # >=1.2.0 upon upgrade of kserve's version
    "mypy>=0.812",
    "pytest-cov",
    "pytest-dependency",
    "pytest-xdist",
    "pytest",
    "recursive-diff>=1.0.0",
    "requests",
    "scikit-learn>=1.1.2",
    "types-python-dateutil",
    "types-PyYAML",
    "types-six",
    "urllib3-mock>=0.3.3",
    "xarray",
    "xgboost==1.6.2",
]

setup(
    name="merlin-sdk",
    version=version,
    description="Python SDK for Merlin",
    url="https://github.com/caraml-dev/merlin",
    author="Merlin",
    packages=find_packages(),
    package_data={"merlin": [
        "docker/pyfunc.Dockerfile", "docker/standard.Dockerfile"]},
    zip_safe=True,
    install_requires=REQUIRES,
    setup_requires=["setuptools_scm"],
    tests_require=TEST_REQUIRES,
    extras_require={'test': TEST_REQUIRES},
    python_requires='>=3.8,<3.11',
    long_description=open("README.md").read(),
    long_description_content_type='text/markdown',
    entry_points='''
        [console_scripts]
        merlin=merlin.merlin:cli
    '''
)



# coding: utf-8

"""
    Merlin

    API Guide for accessing Merlin's model management, deployment, and serving functionalities  # noqa: E501

    OpenAPI spec version: 0.14.0
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

import pprint
import re  # noqa: F401

import six


class StandardTransformerSimulationRequest(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        "payload": "FreeFormObject",
        "headers": "FreeFormObject",
        "config": "FreeFormObject",
        "model_prediction_config": "ModelPredictionConfig",
        "protocol": "Protocol",
    }

    attribute_map = {
        "payload": "payload",
        "headers": "headers",
        "config": "config",
        "model_prediction_config": "model_prediction_config",
        "protocol": "protocol",
    }

    def __init__(
        self,
        payload=None,
        headers=None,
        config=None,
        model_prediction_config=None,
        protocol=None,
    ):  # noqa: E501
        """StandardTransformerSimulationRequest - a model defined in Swagger"""  # noqa: E501
        self._payload = None
        self._headers = None
        self._config = None
        self._model_prediction_config = None
        self._protocol = None
        self.discriminator = None
        if payload is not None:
            self.payload = payload
        if headers is not None:
            self.headers = headers
        if config is not None:
            self.config = config
        if model_prediction_config is not None:
            self.model_prediction_config = model_prediction_config
        if protocol is not None:
            self.protocol = protocol

    @property
    def payload(self):
        """Gets the payload of this StandardTransformerSimulationRequest.  # noqa: E501


        :return: The payload of this StandardTransformerSimulationRequest.  # noqa: E501
        :rtype: FreeFormObject
        """
        return self._payload

    @payload.setter
    def payload(self, payload):
        """Sets the payload of this StandardTransformerSimulationRequest.


        :param payload: The payload of this StandardTransformerSimulationRequest.  # noqa: E501
        :type: FreeFormObject
        """

        self._payload = payload

    @property
    def headers(self):
        """Gets the headers of this StandardTransformerSimulationRequest.  # noqa: E501


        :return: The headers of this StandardTransformerSimulationRequest.  # noqa: E501
        :rtype: FreeFormObject
        """
        return self._headers

    @headers.setter
    def headers(self, headers):
        """Sets the headers of this StandardTransformerSimulationRequest.


        :param headers: The headers of this StandardTransformerSimulationRequest.  # noqa: E501
        :type: FreeFormObject
        """

        self._headers = headers

    @property
    def config(self):
        """Gets the config of this StandardTransformerSimulationRequest.  # noqa: E501


        :return: The config of this StandardTransformerSimulationRequest.  # noqa: E501
        :rtype: FreeFormObject
        """
        return self._config

    @config.setter
    def config(self, config):
        """Sets the config of this StandardTransformerSimulationRequest.


        :param config: The config of this StandardTransformerSimulationRequest.  # noqa: E501
        :type: FreeFormObject
        """

        self._config = config

    @property
    def model_prediction_config(self):
        """Gets the model_prediction_config of this StandardTransformerSimulationRequest.  # noqa: E501


        :return: The model_prediction_config of this StandardTransformerSimulationRequest.  # noqa: E501
        :rtype: ModelPredictionConfig
        """
        return self._model_prediction_config

    @model_prediction_config.setter
    def model_prediction_config(self, model_prediction_config):
        """Sets the model_prediction_config of this StandardTransformerSimulationRequest.


        :param model_prediction_config: The model_prediction_config of this StandardTransformerSimulationRequest.  # noqa: E501
        :type: ModelPredictionConfig
        """

        self._model_prediction_config = model_prediction_config

    @property
    def protocol(self):
        """Gets the protocol of this StandardTransformerSimulationRequest.  # noqa: E501


        :return: The protocol of this StandardTransformerSimulationRequest.  # noqa: E501
        :rtype: Protocol
        """
        return self._protocol

    @protocol.setter
    def protocol(self, protocol):
        """Sets the protocol of this StandardTransformerSimulationRequest.


        :param protocol: The protocol of this StandardTransformerSimulationRequest.  # noqa: E501
        :type: Protocol
        """

        self._protocol = protocol

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(
                    map(lambda x: x.to_dict() if hasattr(x, "to_dict") else x, value)
                )
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(
                    map(
                        lambda item: (item[0], item[1].to_dict())
                        if hasattr(item[1], "to_dict")
                        else item,
                        value.items(),
                    )
                )
            else:
                result[attr] = value
        if issubclass(StandardTransformerSimulationRequest, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, StandardTransformerSimulationRequest):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other
